//-----------------------------------------------------------------------
// <copyright file="ISyncAdvancedSessionOperation.cs" company="Hibernating Rhinos LTD">
//     Copyright (c) Hibernating Rhinos LTD. All rights reserved.
// </copyright>
//-----------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.Linq;

using Raven.Abstractions.Data;
using Raven.Client.Document.Batches;
using Raven.Client.Indexes;

namespace Raven.Client
{
	/// <summary>
	///     Advanced synchronous session operations
	/// </summary>
	public interface ISyncAdvancedSessionOperation : IAdvancedDocumentSessionOperations
	{
		/// <summary>
		///     Access the eager operations
		/// </summary>
		IEagerSessionOperations Eagerly { get; }

		/// <summary>
		///     Access the lazy operations
		/// </summary>
		ILazySessionOperations Lazily { get; }

		/// <summary>
		///     Queries the index specified by <typeparamref name="TIndexCreator" /> using lucene syntax.
		/// </summary>
		/// <typeparam name="T">The result of the query</typeparam>
		/// <typeparam name="TIndexCreator">The type of the index creator.</typeparam>
		IDocumentQuery<T> DocumentQuery<T, TIndexCreator>() where TIndexCreator : AbstractIndexCreationTask, new();

		/// <summary>
		///     Query the specified index using Lucene syntax
		/// </summary>
		/// <param name="indexName">Name of the index.</param>
		/// <param name="isMapReduce">Control how we treat identifier properties in map/reduce indexes</param>
		IDocumentQuery<T> DocumentQuery<T>(string indexName, bool isMapReduce = false);

		/// <summary>
		///     Dynamically query RavenDB using Lucene syntax
		/// </summary>
		IDocumentQuery<T> DocumentQuery<T>();

		/// <summary>
		///     Returns full document url for a given entity
		/// </summary>
		/// <param name="entity">Instance of an entity for which url will be returned</param>
		string GetDocumentUrl(object entity);

		/// <summary>
		///     Loads multiple entities that contain common prefix.
		/// </summary>
		/// <param name="keyPrefix">prefix for which documents should be returned e.g. "products/"</param>
		/// <param name="matches">
		///     pipe ('|') separated values for which document keys (after 'keyPrefix') should be matched ('?'
		///     any single character, '*' any characters)
		/// </param>
		/// <param name="start">number of documents that should be skipped. By default: 0.</param>
		/// <param name="pageSize">maximum number of documents that will be retrieved. By default: 25.</param>
		/// <param name="pagingInformation">used to perform rapid pagination on a server side</param>
		/// <param name="exclude">
		///     pipe ('|') separated values for which document keys (after 'keyPrefix') should not be matched
		///     ('?' any single character, '*' any characters)
		/// </param>
		/// <param name="skipAfter">
		///     skip document fetching until given key is found and return documents after that key (default:
		///     null)
		/// </param>
		T[] LoadStartingWith<T>(string keyPrefix, string matches = null, int start = 0, int pageSize = 25, string exclude = null, RavenPagingInformation pagingInformation = null, string skipAfter = null);

		/// <summary>
		///     Loads multiple entities that contain common prefix and applies specified transformer.
		/// </summary>
		/// <param name="keyPrefix">prefix for which documents should be returned e.g. "products/"</param>
		/// <param name="matches">
		///     pipe ('|') separated values for which document keys (after 'keyPrefix') should be matched ('?'
		///     any single character, '*' any characters)
		/// </param>
		/// <param name="start">number of documents that should be skipped. By default: 0.</param>
		/// <param name="pageSize">maximum number of documents that will be retrieved. By default: 25.</param>
		/// <param name="pagingInformation">used to perform rapid pagination on a server side</param>
		/// <param name="exclude">
		///     pipe ('|') separated values for which document keys (after 'keyPrefix') should not be matched
		///     ('?' any single character, '*' any characters)
		/// </param>
		/// <param name="configure">additional configuration options for operation e.g. AddTransformerParameter</param>
		/// <param name="skipAfter">
		///     skip document fetching until given key is found and return documents after that key (default:
		///     null)
		/// </param>
		TResult[] LoadStartingWith<TTransformer, TResult>(string keyPrefix, string matches = null, int start = 0, int pageSize = 25, string exclude = null, RavenPagingInformation pagingInformation = null, Action<ILoadConfiguration> configure = null, string skipAfter = null) where TTransformer : AbstractTransformerCreationTask, new();

		/// <summary>
		///     Queries the index specified by <typeparamref name="TIndexCreator" /> using lucene syntax.
		/// </summary>
		/// <typeparam name="T">The result of the query</typeparam>
		/// <typeparam name="TIndexCreator">The type of the index creator.</typeparam>
		/// <returns></returns>
		[Obsolete("Use DocumentQuery instead")]
		IDocumentQuery<T> LuceneQuery<T, TIndexCreator>() where TIndexCreator : AbstractIndexCreationTask, new();

		/// <summary>
		///     Query the specified index using Lucene syntax
		/// </summary>
		/// <param name="indexName">Name of the index.</param>
		/// <param name="isMapReduce">Control how we treat identifier properties in map/reduce indexes</param>
		[Obsolete("Use DocumentQuery instead")]
		IDocumentQuery<T> LuceneQuery<T>(string indexName, bool isMapReduce = false);

		/// <summary>
		///     Dynamically query RavenDB using Lucene syntax
		/// </summary>
		[Obsolete("Use DocumentQuery instead")]
		IDocumentQuery<T> LuceneQuery<T>();

		/// <summary>
		///     Sends a multiple faceted queries in a single request and calculates the facet results for each of them
		/// </summary>
		/// <param name="queries">Array of the faceted queries that will be executed on the server-side</param>
		FacetResults[] MultiFacetedSearch(params FacetQuery[] queries);

		/// <summary>
		///     Updates entity with latest changes from server
		/// </summary>
		/// <param name="entity">Instance of an entity that will be refreshed</param>
		void Refresh<T>(T entity);

		/// <summary>
		///     Stream the results on the query to the client, converting them to
		///     CLR types along the way.
		///     <para>Does NOT track the entities in the session, and will not includes changes there when SaveChanges() is called</para>
		/// </summary>
		/// <param name="query">Query to stream results for</param>
		IEnumerator<StreamResult<T>> Stream<T>(IQueryable<T> query);

		/// <summary>
		///     Stream the results on the query to the client, converting them to
		///     CLR types along the way.
		///     <para>Does NOT track the entities in the session, and will not includes changes there when SaveChanges() is called</para>
		/// </summary>
		/// <param name="query">Query to stream results for</param>
		/// <param name="queryHeaderInformation">Information about performed query</param>
		IEnumerator<StreamResult<T>> Stream<T>(IQueryable<T> query, out QueryHeaderInformation queryHeaderInformation);

		/// <summary>
		///     Stream the results on the query to the client, converting them to
		///     CLR types along the way.
		///     <para>Does NOT track the entities in the session, and will not includes changes there when SaveChanges() is called</para>
		/// </summary>
		/// <param name="query">Query to stream results for</param>
		IEnumerator<StreamResult<T>> Stream<T>(IDocumentQuery<T> query);

		/// <summary>
		///     Stream the results on the query to the client, converting them to
		///     CLR types along the way.
		///     <para>Does NOT track the entities in the session, and will not includes changes there when SaveChanges() is called</para>
		/// </summary>
		/// <param name="query">Query to stream results for</param>
		/// <param name="queryHeaderInformation">Information about performed query</param>
		IEnumerator<StreamResult<T>> Stream<T>(IDocumentQuery<T> query, out QueryHeaderInformation queryHeaderInformation);

		/// <summary>
		///     Stream the results of documents search to the client, converting them to CLR types along the way.
		///     <para>Does NOT track the entities in the session, and will not includes changes there when SaveChanges() is called</para>
		/// </summary>
		/// <param name="fromEtag">ETag of a document from which stream should start</param>
		/// <param name="start">number of documents that should be skipped</param>
		/// <param name="pageSize">maximum number of documents that will be retrieved</param>
		/// <param name="pagingInformation">used to perform rapid pagination on a server side</param>
		IEnumerator<StreamResult<T>> Stream<T>(Etag fromEtag, int start = 0, int pageSize = int.MaxValue, RavenPagingInformation pagingInformation = null);

		/// <summary>
		///     Stream the results of documents search to the client, converting them to CLR types along the way.
		///     <para>Does NOT track the entities in the session, and will not includes changes there when SaveChanges() is called</para>
		/// </summary>
		/// <param name="startsWith">prefix for which documents should be returned e.g. "products/"</param>
		/// <param name="matches">
		///     pipe ('|') separated values for which document keys (after 'keyPrefix') should be matched ('?'
		///     any single character, '*' any characters)
		/// </param>
		/// <param name="start">number of documents that should be skipped</param>
		/// <param name="pageSize">maximum number of documents that will be retrieved</param>
		/// <param name="pagingInformation">used to perform rapid pagination on a server side</param>
		/// <param name="skipAfter">
		///     skip document fetching until given key is found and return documents after that key (default:
		///     null)
		/// </param>
		IEnumerator<StreamResult<T>> Stream<T>(string startsWith, string matches = null, int start = 0, int pageSize = int.MaxValue, RavenPagingInformation pagingInformation = null, string skipAfter = null);
	}
}